# -*- coding: utf-8 -*-

import os

import csv
import re
import logging
import luigi
import law
import law.contrib.htcondor
import six

from law.parameter import CSVParameter, NO_STR
from law.util import is_lazy_iterable, make_tuple

from analysis.mixins import HTCondorRemoteStorageMixin


logger = logging.getLogger(__name__)

logger.debug("Loading law contribution wlcg ...")
law.contrib.load("wlcg")


class SpaceSeparatedListParameter(CSVParameter):
    # developed for `law` revision df3d1fd0a15fd73f80eee859a616373d829f4513 (v0.1.12)
    # might need to be changed when updating to newer `law` versions

    def __init__(self, *args, **kwargs):
        # warn that brace expansion is not supported
        if kwargs.pop("brace_expand", False):
            logger.warning("{!r} does not support brace expansion".format(self))
        super(SpaceSeparatedListParameter, self).__init__(*args, **kwargs)

    def _check_len(self, value):
        value = tuple(
            v.replace(",", "\\,")
            if isinstance(v, six.string_types)
            else v
            for v in value
        )
        super(SpaceSeparatedListParameter, self)._check_len(value)

    def parse(self, inp):
        """
        Parse the SpaceSeparatedListParameter, splitting the `inp` string on
        contiguous whitespace.
        """
        if not inp or inp == NO_STR:
            value = tuple()
        elif isinstance(inp, (tuple, list)) or is_lazy_iterable(inp):
            value = make_tuple(inp)
        elif isinstance(inp, six.string_types):
            # replace escaped separators
            if self._escape_sep:
                escaped_sep = "__law_escaped_csv_sep__"
                inp = inp.replace("\\ ", escaped_sep)
            # replace literal commas so they are not parsed
            # by CSV reader
            escaped_comma = "__law_escaped_comma__"
            inp = inp.replace(",", escaped_comma)
            # strip initial and final spaces
            inp = re.sub(r"(^\s+|\s+$)", "", inp)
            # collapse contiguous whitespace to commas
            inp = re.sub(r"\s+", ",", inp)
            # proper csv split
            elems = list(csv.reader([inp]))[0]
            # add back escaped commas per element
            elems = [elem.replace(escaped_comma, ",") for elem in elems]
            # add back escaped separators per element
            if self._escape_sep:
                elems = [elem.replace(escaped_sep, " ") for elem in elems]
            value = tuple(map(self._inst.parse, elems))
        else:
            value = (inp,)

        # apply uniqueness, sort, length and choices checks
        value = self._check_unique(value)
        value = self._check_sort(value)
        self._check_len(value)
        self._check_choices(value)

        return value

    def serialize(self, value):
        """
        Serialize a value into a string that, when parsed as a
        SpaceSeparatedListParameter, yields that value.
        """
        if not value:
            value = tuple()

        value = make_tuple(value)

        # escape any literal spaces in value elements
        value = tuple(
            v.replace(" ", "\\ ")
            if isinstance(v, six.string_types)
            else v
            for v in value
        )

        # apply uniqueness, sort, length and choices checks
        value = self._check_unique(value)
        value = self._check_sort(value)
        self._check_len(value)
        self._check_choices(value)

        return " ".join(str(self._inst.serialize(elem)) for elem in value)



class Task(law.Task):

    name = luigi.Parameter()
    process = luigi.Parameter()

    def local_path(self, *path):
        # check local output env var and raise if not set
        local_output_dir = os.getenv("ANALYSIS_OUTPUT_PATH")
        if not local_output_dir:
            raise RuntimeError(
                "failed to construct local path; environment variable for local "
                "storage ($ANALYSIS_OUTPUT_PATH) is not set",
            )

        return os.path.join(local_output_dir, *path)

    def local_target(self, *path):
        return law.LocalFileTarget(self.local_path(*path))

    def remote_path(self, *path):
        parts = (self.name, self.__class__.__name__) + path
        return os.path.join(*parts)

    def remote_target(self, *path):
        return law.wlcg.WLCGFileTarget(self.remote_path(*path))


class HTCondorJobManager(law.contrib.htcondor.HTCondorJobManager):

    status_line_cre = re.compile("^(\d+\.\d+)" + 4 * "\s+[^\s]+" + "\s+([UIRXSCHE<>])\s+.*$")

    def get_htcondor_version(cls):
        return (8, 6, 5)

    @classmethod
    def map_status(cls, status_flag):
        if status_flag in ("1", "0", "7", "U", "I", "S"):
            return cls.PENDING
        elif status_flag in ("2", "R", "<", ">"):
            return cls.RUNNING
        elif status_flag in ("4", "C"):
            return cls.FINISHED
        elif status_flag in ("5", "H", "E"):
            return cls.FAILED
        else:
            return cls.FAILED


class HTCondorWorkflow(law.contrib.htcondor.HTCondorWorkflow, HTCondorRemoteStorageMixin):

    # Luigi
    bootstrap_file = luigi.Parameter()
    base_runcard = luigi.Parameter()
    steer_dir = luigi.Parameter()
    output_collection_cls = law.SiblingFileCollection
    batchlocation = luigi.Parameter(default="ETP")

    # HTCondor
    htcondor_accounting_group = luigi.Parameter()
    htcondor_docker_image = luigi.Parameter()
    htcondor_request_memory = luigi.Parameter()
    htcondor_requirements = luigi.Parameter()
    htcondor_user_proxy = luigi.Parameter()
    htcondor_remote_job = luigi.BoolParameter(default=True)
    htcondor_request_cpus = luigi.Parameter()
    htcondor_use_cpus = luigi.Parameter()
    htcondor_universe = luigi.Parameter()
    htcondor_walltime = luigi.Parameter()
    # Set default for new luigi parameter htcondor_request_disk
    htcondor_request_disk = luigi.IntParameter(default=2000000)

    def htcondor_create_job_manager(self, **kwargs):
        return HTCondorJobManager()

    def htcondor_output_directory(self):
        if self.htcondor_remote_storage:
            logger.debug("HTCondor output: Remote storage requested: %s", self.htcondor_remote_storage)
            return law.wlcg.WLCGDirectoryTarget(self.remote_path())
        else:
            logger.debug("HTCondor output: No remote storage requested: %s", self.htcondor_remote_storage)
            return law.LocalDirectoryTarget(self.local_path())

    def htcondor_create_job_file_factory(self):
        factory = super(HTCondorWorkflow, self).htcondor_create_job_file_factory()
        factory.is_tmp = False
        return factory

    def htcondor_bootstrap_file(self):
        return law.util.rel_path(__file__, self.bootstrap_file)

    def htcondor_job_config(self, config, job_num, branches):
        config.custom_content = []
        # Do NOT use getenv at ETP!
        #config.custom_content.append(("getenv", "true"))
        config.render_variables["analysis_path"] = os.getenv("ANALYSIS_PATH")
        config.custom_content.append(("accounting_group", self.htcondor_accounting_group))
        config.custom_content.append(("docker_image", self.htcondor_docker_image))
        config.custom_content.append(("RequestMemory", self.htcondor_request_memory))
        config.custom_content.append(("Requirements", self.htcondor_requirements))
        config.custom_content.append(("x509userproxy", self.htcondor_user_proxy))
        config.custom_content.append(("+RemoteJob", self.htcondor_remote_job))
        config.custom_content.append(("request_cpus", self.htcondor_request_cpus))
        config.custom_content.append(("universe", self.htcondor_universe))
        config.custom_content.append(("+RequestWalltime", self.htcondor_walltime))
        config.custom_content.append(("RequestDisk", self.htcondor_request_disk))

        prevdir = os.getcwd()
        os.system('cd $ANALYSIS_PATH')
        if os.path.isfile('analysis.tar.gz'):
            os.remove('analysis.tar.gz')
        os.system('tar --exclude=luigi/.git --exclude={}/SteerModify.str -czf analysis.tar.gz analysis luigi.cfg law.cfg luigi law six enum34-1.1.10 python-dateutil-2.8.2 {} {}'.format(self.steer_dir, self.base_runcard, self.steer_dir))
        os.chdir(prevdir)

        config.input_files['analysis'] = law.util.rel_path(__file__, '../analysis.tar.gz')
        return config

    def htcondor_use_local_scheduler(self):
        return True
